package pt.utl.ist.online.learning;

import java.io.Serializable;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;




import pt.utl.ist.online.learning.engines.KernelPegasosEngine;
import pt.utl.ist.online.learning.engines.OnlineEngine;
import pt.utl.ist.online.learning.engines.info.EngineObserver;
import pt.utl.ist.online.learning.exceptions.InvalidVectorIndexException;
import pt.utl.ist.online.learning.kernels.Kernel;
import pt.utl.ist.online.learning.utils.DataObject;


/**
 * Runs the updateModel method of an OnlineEngine for a list of data samples 
 * and the list of the associated binary classes. 
 * <p>
 * The objective of this class is to train or retrain an OnlineEngine 
 * using a set of data instances. It iteratively calls the updateModel method 
 * until the number of iterations reach a number specified by variable 
 * numEpochs over the list of data samples
 *
 * @author Goncalo Simoes
 * @author Joao L. M. Pereira
 * @version 0.1
 * @since 2015-06-18
 */

public class KernelBinaryOnlineAlgorithm<E> implements Serializable {

	private OnlineEngine<E> engine;
	private int randomNumber = 5443;

	/**
	 * Calls the training for the OnlineEngine
	 * 
	 * @param trainingData
	 *            the list of data samples used for training
	 * @param labels
	 *            a Map containing the data samples as keys and their labels as
	 *             values       
	 * @param currentEpoch
	 *            the initial iteration of the algorithm, 
	 *            used mainly to continue a already trained OnlineEngine
	 * @param numEpochs
	 * 			  the number of iterations to go over the trainingData
	 * @param engine
	 *            the OnlineEngine to be trained
	 * @param cacheMatrix
	 *            boolean to request a pre-computed matrix in kernel based 
	 *            OnlineLearning
	 * @param observers
	 *            A list of EngineObservers that will obtain information 
	 *            during the training
	 * @param classifier
	 *            the name of the classifier, used mainly to distinguish 
	 *            binary engines in multi-class problems
	 * @throws InvalidVectorIndexException when a DataObject from the data 
	 *            instances to be transforms has occupied reserved space for 
	 *            constants
	 */

	public KernelBinaryOnlineAlgorithm(List<DataObject<E>> trainingData,
			Map<Integer, Boolean> labels, int currentEpoch, int numEpochs,
			OnlineEngine<E> engine, boolean cacheMatrix,
			List<EngineObserver<E>> observers, String classifier) throws InvalidVectorIndexException{		

		process(trainingData,labels, currentEpoch, numEpochs,engine,cacheMatrix,observers,classifier);
	}

	/**
	 * Calls the training for the OnlineEngine
	 * 
	 * @param trainingData
	 *            the list of data samples used for training
	 * @param labels
	 *            a Map containing the data samples as keys and their labels as
	 *             values 
	 * @param numEpochs
	 * 			  the number of iterations to go over the trainingData
	 * @param engine
	 *            the OnlineEngine to be trained
	 * @param cacheMatrix
	 *            boolean to request a pre-computed matrix in kernel based 
	 *            OnlineLearning
	 * @param observers
	 *            A list of EngineObservers that will obtain information 
	 *            during the training
	 * @param classifier
	 *            the name of the classifier, used mainly to distinguish 
	 *            binary engines in multi-class problems
	 * @throws InvalidVectorIndexException when a DataObject from the data 
	 *            instances to be transforms has occupied reserved space for 
	 *            constants
	 */
	public KernelBinaryOnlineAlgorithm(List<DataObject<E>> trainingData,
			Map<Integer, Boolean> labels, int numEpochs,
			OnlineEngine<E> engine, boolean cacheMatrix,
			List<EngineObserver<E>> observers, String classifier) throws InvalidVectorIndexException{		

		process(trainingData,labels,numEpochs,engine,cacheMatrix,observers,classifier);
	}

	private void process(List<DataObject<E>> trainingData,
			Map<Integer, Boolean> labels, int numEpochs,
			OnlineEngine<E> engine, boolean cacheMatrix,
			List<EngineObserver<E>> observers, String classifier) throws InvalidVectorIndexException{
		process(trainingData,labels,0,numEpochs,engine,cacheMatrix,observers,classifier);
	}

	private void process(List<DataObject<E>> trainingData,
			Map<Integer, Boolean> labels, int inicialEpoch, int numEpochs,
			OnlineEngine<E> engine, boolean cacheMatrix,
			List<EngineObserver<E>> observers, String classifier) throws InvalidVectorIndexException {


		this.engine = engine;

		Collections.sort(trainingData);

		Kernel<E> kernel = null;

		if(cacheMatrix)
			if(engine instanceof KernelPegasosEngine){
				kernel = ((KernelPegasosEngine<E>) engine).getKernel();
				Kernel<E> cacheKernel = getComputedKernels(kernel,trainingData);
				((KernelPegasosEngine<E>) engine).setKernel(cacheKernel);
			}else
				new UnsupportedOperationException("A cache matrix for kernels is not available for " + engine.getClass());


		Random r = new Random(randomNumber);

		engine.start();

		if(inicialEpoch != 0){
			for(int i = 0; i < inicialEpoch; i++ )
				Collections.shuffle(trainingData, r);
		}

		for(int currentEpoch=inicialEpoch + 1; currentEpoch<=numEpochs; currentEpoch++){
			int updatesCount=0;
			int elem=0;


			Collections.shuffle(trainingData, r);
			for(DataObject<E> data : trainingData){


				Boolean desiredOutput = labels.get(data.getId());
				boolean updated = engine.updateModel(data, desiredOutput);
				if(updated){
					updatesCount++;
				}

				elem++;

			}

			if(currentEpoch%10==0){
				Double objvalue = null;
				if(engine instanceof KernelPegasosEngine){
					objvalue = ((KernelPegasosEngine<E>) engine).objectiveValue(trainingData, labels);
				}else{
					throw new UnsupportedOperationException();
				}
				for(EngineObserver<E> o : observers)
					o.notify(engine, classifier, currentEpoch, updatesCount, objvalue);
			}
			else{
				for(EngineObserver<E> o : observers)
					o.notify(engine, classifier, currentEpoch, updatesCount);
			}



		}
		for(EngineObserver<E> o : observers)
			o.close(engine, classifier);

		if(cacheMatrix){
			if(engine instanceof KernelPegasosEngine)
				((KernelPegasosEngine<E>) engine).setKernel(kernel);
			else
				new UnsupportedOperationException("A cache matrix for kernels is not available for " + engine.getClass());
		}
		engine.compileModel();	

	}



	private Kernel<E> getComputedKernels(Kernel<E> notComputedKernel, 
			List<DataObject<E>> trainingData) {

		Collections.sort(trainingData);

		Map<Integer,Integer> conversorId = new HashMap<Integer,Integer>(trainingData.size());
		int j = 0;
		for(DataObject<E> sample : trainingData){
			conversorId.put(sample.getId(), j);
			j++;
		}

		Kernel<E> onlineKernel = notComputedKernel.computeKernelMatrix(trainingData);

		return onlineKernel;
	}


	/**
	 * Assigns a binary label to a data sample
	 * 
	 * @param vector the data sample
	 * @return the binary label
	 * @throws InvalidVectorIndexException indicates that the DataObject that 
	 * this method tried to convert contains a reserved space for a constant
	 * that it is occupied
	 */
	public boolean classify(DataObject<E> vector) throws InvalidVectorIndexException{
		vector = engine.convertVector(vector);
		return engine.getResult(vector);
	}

	/**
	 * Trains the OnlineEngine with a data sample
	 * 
	 * @param inputVector
	 *            the data sample
	 * @param desiredOutput
	 *            the binary label
	 * @throws InvalidVectorIndexException indicates that the DataObject that 
	 * this method tried to convert contains a reserved space for a constant
	 * that it is occupied
	 */        

	public void addExampleAndUpdate(DataObject<E> inputVector, boolean desiredOutput) throws InvalidVectorIndexException{
		inputVector = engine.convertVector(inputVector);
		engine.updateModel(inputVector, desiredOutput);
	}


	/**
	 * Used for SVM based OnlineEngine, it returns the weights for the Support 
	 * Vectors
	 * 
	 * @return the Map containing the Support Vectors and the weights
	 */   
	public Map<DataObject<E>,Double> getWeights(){
		if(engine instanceof KernelPegasosEngine){
			return ((KernelPegasosEngine) engine).getWeights();
		}
		throw new UnsupportedOperationException("getWeights() for " + engine.getClass());
	}

	/**
	 * Gets the bias term value present at SVM based OnlineEngine
	 * 
	 * @return the value of the bias term
	 */   
	public double getB() {
		if(engine instanceof KernelPegasosEngine){
			return ((KernelPegasosEngine) engine).getB();
		}
		throw new UnsupportedOperationException("getB() for " + engine.getClass());
	}

	private Map<Integer,DataObject<E>> convertTrainingData(Map<Integer,DataObject<E>> trainingData) throws InvalidVectorIndexException{
		Map<Integer,DataObject<E>> result = new HashMap<Integer, DataObject<E>>();

		for(Entry<Integer,DataObject<E>> entry : trainingData.entrySet())
			result.put(entry.getKey(), engine.convertVector(entry.getValue()));


		return result;
	}



}
