package pt.utl.ist.online.learning.engines;

import java.io.Serializable;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;


import pt.utl.ist.online.learning.exceptions.InvalidVectorIndexException;
import pt.utl.ist.online.learning.kernels.Kernel;
import pt.utl.ist.online.learning.utils.DataObject;

/**
 * implements the Pegasos algorithm and maintains the SVM model produced by 
 * the algorithm in such a way it is allowed to be re-trained
 * <P>
 *The pegasos algorithm is present in the paper Pegasos: Primal Estimated 
 *sub-GrAdient SOlver for SVM. by Shalev-Shwartz, Shai, Yoram Singer,
 * and Nathan Srebro. This class is an implementation of an adaptation of
 *  Pegasos to be used with a kernel function.
 *
 * @author Goncalo Simoes
 * @author Joao L. M. Pereira
 * @version 0.1
 * @since 2015-06-18
 *
 * @param <E> the class of the data objects used in the model
 */

public class KernelPegasosEngine<E> implements OnlineEngine<E>, Serializable {

	/**
	 * Defines the configuration parameters to create OnlineEngines of the
	 *  type <code>KernelPegasosEngine</code>
	 *
	 * @author Joao L. M. Pereira
	 * @version 0.1
	 * @since 2015-06-18
	 *
	 * @param <E1> the class of the data objects used in the model
	 */

	public static class Conf<E1> implements OnlineEngine.Conf<E1>{

		double lambda;
		double hingeParameter;
		Kernel<E1> k = null;

		/**
		 * Creates a Conf object that can create objects of 
		 * <code>KernelPegasosEngine</code>
		 * 
		 * @param lambda a double containing the lambda parameter of the SVM 
		 * model to be created
		 * @param hingeParameter parameter that forces more or less precision. 
		 * The recommended and the usual is to set this parameter to zero
		 * @param k the kernel used to compare pairs of data instances
		 */
		public Conf(double lambda,
				double hingeParameter, Kernel<E1> k){
			this.lambda = lambda;
			this.hingeParameter = hingeParameter;
			this.k = k;
		}

		@Override
		public OnlineEngine<E1> createEngine() {
			return new KernelPegasosEngine<E1>(lambda,hingeParameter,k);
		}

		@Override
		public void cacheMatrix(List<DataObject<E1>> trainingData) {
			k = getComputedKernel(k,trainingData);
		}

		private Kernel<E1> getComputedKernel(Kernel<E1> notComputedKernel, 
				List<DataObject<E1>> trainingData) {

			Collections.sort(trainingData);
			int dataSize = trainingData.size();
			DataObject<E1> lastElementValue = trainingData.get(dataSize - 1);

			if((dataSize -1) ==  lastElementValue.getId()){
				Kernel<E1> onlineKernel = notComputedKernel.computeKernelMatrix(trainingData);
				return onlineKernel;
			}

			Map<Integer,Integer> conversorId = new HashMap<Integer,Integer>(trainingData.size());
			int j = 0;
			for(DataObject<E1> sample : trainingData){
				conversorId.put(sample.getId(), j);
				j++;
			}

			Kernel<E1> onlineKernel = notComputedKernel.computeKernelMatrix(trainingData, conversorId);
			return onlineKernel;
		}
	}



	private Map<DataObject<E>,Double> weights;
	private double b=0.0;
	private double currentNormSquared=0;
	private double lambda;
	private double hingeParameter;
	private Kernel<E> kernel;
	private double t=0;

	private long kernelAccesses = 0;

	/**
	 * Creates a Pegasos Engine that use a Kernel
	 * 
	 * @param lambda a double containing the lambda parameter of the SVM 
	 * model to be created
	 * @param hingeParameter parameter that forces more or less precision. 
	 * The recommended and the usual is to set this parameter to zero
	 * @param k the kernel used to compare pairs of data instances
	 */
	public KernelPegasosEngine(double lambda, double hingeParameter, Kernel<E> k) {
		this.lambda=lambda;
		weights = new HashMap<DataObject<E>, Double>();
		this.kernel=k;
		this.hingeParameter=hingeParameter;
	}


	@Override
	public boolean updateModel(DataObject<E> inputVector, boolean desiredOutput){

		boolean update = false;
		t++;
		double y=1.0;
		if(!desiredOutput){
			y=-1.0;
		}
		double inner = decisionFunction(inputVector);
		kernelAccesses += weights.size();

		reWeight(1.0-(1/t));

		if(y*(inner-b)<=hingeParameter){
			Double alpha = weights.get(inputVector);
			if(alpha==null){
				alpha=0.0;
			}

			double diff=y/(t*lambda); 
			alpha+=diff;
			b-=diff;

			currentNormSquared+=((2*diff*inner*(1.0-(1/t)))+Math.pow(diff, 2)*kernel.compute(inputVector, inputVector));
			kernelAccesses++;

			weights.put(inputVector, alpha);

			update = true;
		}

		double norm=Math.sqrt(currentNormSquared);
		double aux = Math.min(1.0, 1.0/(norm*Math.sqrt(lambda)));
		reWeight(aux);
		b *= aux;

		return update;
	}

	/**
	 * Adds a new array of data instance to this online model.
	 * In other words it re-trains this model using the new array of data instance.
	 * 
	 * <p>
	 * 
	 * This method is intended for batch processing
	 * 
	 * @param inputVectors the array of data instance of type E encapsulated by a DataObject
	 * @param desiredOutputs an array of booleans that indicate the true label for the data
	 * instances at <code>inputVectors</code>
	 * @return true if this model was updated and false otherwise
	 */
	public boolean updateModel(DataObject<E> inputVectors[], boolean desiredOutputs[]){
		int k = inputVectors.length;

		t++;

		List<DataObject<E>> nonZeroLoss = new ArrayList<DataObject<E>>();
		List<Double> nonZeroLossAlphas = new ArrayList<Double>();
		List<Boolean> nonZeroLossYs = new ArrayList<Boolean>();




		double learningRateK = 1/(t*lambda*k);
		double acumulatorSVsAllvsNew = 0;

		double acumulatorSVsNewvsNew = 0;

		double acumulatorSelfKernels = 0;

		for (int i = 0; i<k; i++){
			DataObject<E> inputVector = inputVectors[i];
			boolean desiredOutput = desiredOutputs[i];

			double y=1.0;
			if(!desiredOutput){
				y=-1.0;
			}
			double inner = decisionFunction(inputVector);


			if(y*(inner-b)<=hingeParameter){

				double diff=y*learningRateK; 
				b-=diff;
				acumulatorSVsAllvsNew += inner * (1 - (1/t)) * y;

				for( int j=0;j< nonZeroLoss.size(); j++){
					DataObject<E> sv = nonZeroLoss.get(j);
					int ySv;
					if(nonZeroLossYs.get(j)){
						ySv = 1;
					}else{
						ySv = -1;
					}
					acumulatorSVsNewvsNew += ySv * y * kernel.compute(inputVector, sv);	
				}

				acumulatorSelfKernels += kernel.compute(inputVector, inputVector);

				nonZeroLoss.add(inputVector);
				nonZeroLossAlphas.add(diff);
				nonZeroLossYs.add(desiredOutput);
			}

		}
		reWeight(1.0-(1/t));

		currentNormSquared+= (2 * learningRateK * acumulatorSVsAllvsNew) 
				+ (Math.pow(learningRateK, 2) * (2 * acumulatorSVsNewvsNew + acumulatorSelfKernels));

		for( int j=0;j< nonZeroLoss.size(); j++){
			DataObject<E> inputVector = nonZeroLoss.get(j);
			Double diff = nonZeroLossAlphas.get(j);

			Double alpha = weights.get(inputVector);
			if(alpha==null){
				alpha=0.0;
			}

			weights.put(inputVector, (alpha + diff));
		}



		double norm=Math.sqrt(currentNormSquared);
		double aux = Math.min(1.0, 1.0/(norm*Math.sqrt(lambda)));
		reWeight(aux);
		b *= aux;

		return true;
	}

	private void reWeight(double d) {
		currentNormSquared*=(d*d);
		for(Entry<DataObject<E>,Double> supportVector : weights.entrySet()){
			DataObject<E> sv = supportVector.getKey();
			weights.put(sv,supportVector.getValue()*d);
		}
	}

	@Override
	public boolean getResult(DataObject<E> vector){
		double currentResult = decisionFunction(vector)-b;
		//kernel.clear();
		if(currentResult>0){
			return true;
		}else{
			return false;
		}
	}

	@Override
	public double getScore(DataObject<E> vector) {
		return decisionFunction(vector)-b;
	}

	private double decisionFunction(DataObject<E> x){
		double sum = 0.0;
		for(Entry<DataObject<E>,Double> supportVector : weights.entrySet()){
			DataObject<E> sv = supportVector.getKey();
			double weight = supportVector.getValue();

			sum+= weight*kernel.compute(sv, x);
		}
		return sum;
	}

	@Override
	public DataObject<E> convertVector(DataObject<E> x) throws InvalidVectorIndexException{
		return x;
	}

	/**
	 * Gets the kernel used in this engine
	 * 
	 * @return the kernel object for data of type E
	 */
	public Kernel<E> getKernel(){
		return kernel;
	}

	/**
	 * Sets a new kernel object. This method is usually used to set a new 
	 * kernel based on a new catch but that the result is the same as the 
	 * old kernel
	 * 
	 * @param kernel the new kernel that will be used in this model
	 */
	public void setKernel(Kernel<E> kernel){
		this.kernel = kernel;
	}

	/**
	 * Gets the number of SVs (Support Vectors)
	 * SVs are data instance with non-zero weights in this model
	 * 
	 * @return the number of SVs
	 */
	public int numSVs() {
		return weights.size();
	}

	@Override
	public void compileModel() {
		kernel.clear();
		kernel.setUseCache(false);
	}

	@Override
	public void start() {
		kernel.setUseCache(true);
	}

	/**
	 * Gets the number of calls made to the kernels during the training
	 * 
	 * @return a long containing the number of kernel accesses during training
	 */
	public long getKernelAccesses(){
		return kernelAccesses;
	}

	/**
	 * Gets the weights for each SV (Support Vector) in this model
	 * SV is a data instance with non-zero weight, it is used to compose this model
	 * 
	 * @return a Map containing as key, each SVs, and as value, the weight of the 
	 * SV in this model
	 */
	public Map<DataObject<E>,Double> getWeights(){
		return weights;
	}

	/**
	 * Gets the bias term used in this model
	 * 
	 * @return the bias term value
	 */
	public double getB(){
		return b;
	}


	/**
	 * Computes the objective function value for the provided labelled data 
	 * samples
	 * <p>
	 * 
	 * This method is manly used with all the training data samples to obtained
	 *  a convergence value during the training iterations
	 * 
	 * @param trainingData the list of data samples used to compute the objective function value
	 * @param desiredLabels a map containing the labels for the <code>trainingData</code>
	 * @return the objective function value
	 */
	public double objectiveValue(List<DataObject<E>> trainingData, Map<Integer, Boolean> desiredLabels) {

		double sumSamples = 0;
		for( DataObject<E> data : trainingData){
			int id = data.getId();

			double score = getScore(data);
			Boolean trueLabel = desiredLabels.get(id);
			if(trueLabel){
				score *= 1;
			}else{
				score *= -1;
			}

			sumSamples += Math.max(0.0, 1 - score);
		}


		double objectivFunctionResult = (lambda/2) * currentNormSquared + (sumSamples/trainingData.size());

		return objectivFunctionResult;
	}






}
