package pt.utl.ist.online.learning.engines;

import java.io.Serializable;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import pt.utl.ist.online.learning.exceptions.InvalidVectorIndexException;
import pt.utl.ist.online.learning.utils.DataObject;



/**
 * Specifies the behaviour of any online engine.
 * Online engine is a machine learning model that can be retained so apart
 * from storing the model's information it complements with the algorithm for
 * retraining through the use of its methods.
 *
 * @author Goncalo Simoes
 * @author Joao L. M. Pereira
 * @version 0.1
 * @since 2015-06-18
 *
 * @param <E> the class of the data objects used in the model
 */
public interface OnlineEngine<E> extends Serializable {

	/**
	 * The interface OnlineEngine.Conf specifies the parameters to
	 * create an OnlineEngine object, this class is mainly intended to define
	 * the behaviour of an OnlineEngine and reproduce it for solutions that
	 * need to replicate OnlineEngines with the same initial properties. 
	 * <P>
	 * An example is to solve a multi-class problem with binary classifiers,
	 * where we need to use an heuristic (e.g., One-vs-all)
	 * 
	 * 
	 * @author Joao L. M. Pereira
	 * @version 0.1
	 * @since 2015-06-18
	 * @param <E1> the class of the data objects used in this OnlineEngine
	 * @see OnlineEngine
	 */
	public interface Conf<E1> extends Serializable {

		/**
		 * Creates an OnlineEngine based on the parameters provided when the 
		 * Conf object was created
		 * 
		 * @return an OnlineEngine object that uses the parameters set in this
		 *  Conf
		 */
		public OnlineEngine<E1> createEngine();

		/**
		 * Creates a Cache based on a matrix stored memory for Kernels
		 * This type of cache occupies less memory and it provides fast access,
		 * the draw back is that to create this matrix, the method computes
		 * the kernel for every pair of data instance
		 * 
		 * @param trainingData the training data instances used to create the
		 *  matrix
		 */
		public void cacheMatrix(List<DataObject<E1>> trainingData);
	}

	/**
	 * Adds a new data instance to this online model.
	 * In other words it re-trains this model using the new data instance.
	 * 
	 * @param inputVector the data instance of type E encapsulated by a DataObject
	 * @param desiredOutput a boolean indicating the true label for the data
	 * instance at inputVector
	 * @return true if this model was updated and false otherwise
	 */
	public boolean updateModel(DataObject<E> inputVector, boolean desiredOutput);

	/**
	 * Uses this online model to predict the label for a data 
	 * instance of type E encapsulated by a DataObject
	 * 
	 * @param vector an input DataObject to be predicted
	 * @return a boolean label assigned to the DataObject
	 */
	public boolean getResult(DataObject<E> vector);

	/**
	 * Transforms a DataObject into another DataObject with modifications 
	 * to be in the right format for this engine 
	 * @param x a DataObject that encapsulates the data instance to be 
	 * transformed
	 * @return a new data instance inside of a DataObject
	 * @throws InvalidVectorIndexException when the DataObject to be transformed 
	 * has occupied reserved constants
	 */
	public DataObject<E> convertVector(DataObject<E> x) throws InvalidVectorIndexException;

	/**
	 * Calls the finishing procedures after the training of this Online Engine.
	 * <p>
	 * compileModel is mainly intended to reduce the model size, that is compile the 
	 * model
	 * 
	 */
	public void compileModel();


	/**
	 * Calls the initial procedures, before starting using the Online Engine for training.
	 * 
	 */
	public void start();

	/**
	 * Gets the certainty score of this model prediction for a data instance of type E
	 * 
	 * @param vector the data instance that will it receive the score
	 * @return A score value that can be a positive number if the certainty is for the true label and false for the negative label. 
	 * This model is more uncertain about the data instance label as close this returned value is to zero
	 */
	public double getScore(DataObject<E> vector);
}
