package pt.utl.ist.online.learning.kernels;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.AbstractMap.SimpleEntry;
import java.util.concurrent.ConcurrentMap;

import pt.utl.ist.online.learning.exceptions.InvalidVectorIndexException;
import pt.utl.ist.online.learning.utils.DataObject;

import com.google.common.base.Function;
import com.google.common.collect.MapMaker;


/**
 * Specifies the generic behaviour of a kernel. It includes kernel
 * values normalisation and caches.
 * 
 * @author Goncalo Simoes
 * @author Joao L. M. Pereira
 * @version 0.1
 * @since 2015-06-18
 *
 * @param <E> the class of the data instances to be compared by this kernel
 */
public abstract class Kernel<E> implements Serializable{

	transient private Map<Integer,Double> m_mapStoK;

	protected boolean bNorm = true;

	private boolean cacheNorm = true;

	private boolean mapCache = false;

	transient private ConcurrentMap<Entry<DataObject<E>, DataObject<E>>, 
	Double> kernelCache;



	/**
	 * Creates a kernel object the returned kernel value for the computation
	 *  can be normalised and can or not use of caches
	 * 
	 * @param norm a boolean that tells the kernel to return the normalised values
	 * @param cacheNorm a boolean that controls the use of a cache for values 
	 * returned by computing the kernel for two equal data instances
	 * @param mapCache a boolean that allows the use of a cache based on Maps for 
	 * kernel values
	 */
	public Kernel(boolean norm, boolean cacheNorm, boolean mapCache){
		this.bNorm = norm;
		this.mapCache = mapCache;
		if(bNorm && cacheNorm){
			m_mapStoK = new HashMap<Integer,Double>();
		}
	}


	public void setUseCache(boolean useCache) {
		if(useCache){
			if(mapCache)
				createMapCache();

			if(bNorm && cacheNorm && m_mapStoK == null)
				m_mapStoK = new HashMap<Integer,Double>();
		}else{
			mapCache = false;
			this.clear();
		}

	}

	private void createMapCache(){
		if(kernelCache == null)
			kernelCache = new MapMaker()
		//	.initialCapacity(100000)
		//.softKeys()
		.softValues()
		.makeComputingMap(
				new Function<Entry<DataObject<E>, DataObject<E>>, Double>() {
					public Double apply(Entry<DataObject<E>, DataObject<E>> key) {
						DataObject<E> e1 = key.getKey();
						DataObject<E> e2 = key.getValue();
						return computeWithoutMapCache(e1,e2);
					}
				});
	}

	/**
	 * Computes the kernel value for two data instances
	 * 
	 * @param x1 a data instance
	 * @param x2 a data instance
	 * @return a double containing the value computed by the kernel
	 */
	public double compute(DataObject<E> x1, DataObject<E> x2){
		if(mapCache){
			return computeWithMapCache(x1,x2);
		}else{
			return computeWithoutMapCache(x1,x2);
		}
	}

	private double computeWithMapCache(DataObject<E> x1, DataObject<E> x2){
		int id1 = x1.getId();
		int id2 = x2.getId();
		if(id1>id2){
			int temp = id1;
			id1=id2;
			id2=temp;

			DataObject<E> tempV = x1;
			x1=x2;
			x2=tempV;
		}
		Entry<DataObject<E>, DataObject<E>> e = new SimpleEntry<DataObject<E>, DataObject<E>>(x1,x2);
		return kernelCache.get(e);

	}

	private double computeWithoutMapCache(DataObject<E> x1, DataObject<E> x2){
		if(bNorm){
			return computeWithNorm(x1, x2);
		}else{
			return computeValue(x1, x2);
		}
	}

	private double computeWithNorm(DataObject<E> x1, DataObject<E> x2){

		int id1 = x1.getId();
		int id2 = x2.getId();
		if(id1>id2){
			int temp = id1;
			id1=id2;
			id2=temp;

			DataObject<E> tempV = x1;
			x1=x2;
			x2=tempV;
		}

		double k1 = selfKernel(x1);
		double k2 = selfKernel(x2);

		double k = computeValue(x1, x2);
		if (k == 0)
			return 0;

		// normalize
		return k / Math.sqrt (k1 * k2);	
	}

	private double selfKernel(DataObject<E> s)
	{

		if(cacheNorm == false || s.getId() < 0){
			return computeValue(s, s);
		}
		int id = s.getId();
		Double dblk = (Double) m_mapStoK.get(id);
		if (dblk == null) {
			double k = computeValue(s, s);
			m_mapStoK.put(id, k);
			return k;
		}

		return dblk.doubleValue();
	}

	protected abstract double computeValue(DataObject<E> v1, DataObject<E> v2);

	/**
	 * Transforms a DataObject into another DataObject with modifications 
	 * to be in the right format for this kernel 
	 * @param x a DataObject that encapsulates the data instance to be 
	 * transformed
	 * @return a new data instance inside of a DataObject
	 * @throws InvalidVectorIndexException when the DataObject to be transformed 
	 * has occupied reserved constants
	 */
	public abstract DataObject<E> convertVector(DataObject<E> x) throws InvalidVectorIndexException;


	public void clear() {
		if(kernelCache != null){
			kernelCache.clear();
		}
	}



	/**
	 * Creates a matrix to be used for caching purposes.
	 * It computes every pair of data instances contained in the sample.
	 * This method also uses an already provided conversor that maps 
	 * the id of a data instance to a position in the matrix
	 * 
	 * @param samples the data instances
	 * @param conversorId the conversor 
	 * @return a matrix containing the kernel values for all the possible 
	 * pairs of the provided samples
	 */
	public Kernel<E> computeKernelMatrix(List<DataObject<E>> samples, Map<Integer,Integer> conversorId){
		bNorm = true;
		int na = samples.size(); // Dimension of the matrix
		int nv = na * (na - 1 ) / 2; // 1 + 2 + 3 + .. + na
		double[] v = new double[nv];

		int k = 0; // index in v.
		for (int i = 0; i < na; ++i) {

			DataObject<E> sample1 = samples.get(i);


			for (int j = 0; j < i; ++j) {
				DataObject<E> sample2 = samples.get(j);

				v[k++] = computeWithoutMapCache(sample1,sample2);

			}
		}


		System.out.println("Kernel Matrix Computed! " + this.toString());
		m_mapStoK.clear();
		m_mapStoK = null;

		return new ComputedKernel<E>(v, conversorId, this);


	}

	/**
	 * Creates a matrix to be used for caching purposes.
	 * It computes every pair of data instances contained in the sample.
	 * 
	 * @param samples the data instances
	 * @return a matrix containing the kernel values for all the possible 
	 * pairs of the provided samples
	 */
	public Kernel<E> computeKernelMatrix(List<DataObject<E>> samples){

		int na = samples.size(); // Dimension of the matrix
		int nv = na * (na - 1 ) / 2; // 1 + 2 + 3 + .. + na
		double[] v = new double[nv];

		int k = 0; // index in v.
		for (int i = 0; i < na; ++i) {

			DataObject<E> sample1 = samples.get(i);
			for (int j = 0; j < i; ++j) {
				DataObject<E> sample2 = samples.get(j);
				v[k++] = computeWithoutMapCache(sample1,sample2);
			}
		}


		System.out.println("Kernel Matrix Computed! " + this.toString());
		m_mapStoK.clear();
		m_mapStoK = null;

		return new ComputedKernel<E>(v, null, this);
	}

	private void writeObject(ObjectOutputStream oos) throws IOException {
		kernelCache.clear();
		oos.defaultWriteObject();
	}

	private void readObject(ObjectInputStream ois) throws IOException,
	ClassNotFoundException {
		ois.defaultReadObject();
		if(bNorm && cacheNorm){
			m_mapStoK = new HashMap<Integer,Double>();
		}
	}
	
	/**
	 * Get the bNorm variable which tells if the kernel is normalised
	 * 
	 * @return true if the kernel is normalised and false otherwise
	 */
	public boolean getBNorm(){
		return bNorm;
	}
}
