package pt.utl.ist.online.learning.multiclass;

import java.io.Serializable;
import java.util.List;
import pt.utl.ist.online.learning.engines.OnlineEngine;
import pt.utl.ist.online.learning.engines.info.EngineObserver;
import pt.utl.ist.online.learning.utils.DataObject;

/**
 * Implements the abstract class which describes a MultiClass method using 
 * OnlineEngines. It is used to train and classify for multi-class problems.
 * In order to create OnlineEngines, this class receives an OnlineEngine.Conf 
 * object
 * 
 * @author Joao L. M. Pereira
 * @version 0.1
 * @since 2015-06-18
 * 
 * @param <E> the class of the data sample
 * @param <L> the class of the labels for each data sample
 */
public abstract class MulticlassMethod<E,L extends Comparable<L>> implements Serializable {
	protected OnlineEngine.Conf<E> conf;

	/**
	 * Creates a multi-class method by storing the OnlineEngine.Conf
	 * 
	 * @param conf the OnlineEngine.Conf object that stores the parameters 
	 * and can create OnlineEngines
	 */
	public MulticlassMethod(OnlineEngine.Conf<E> conf){
		this.conf = conf;
	}
	
	/**
	 * Trains the multi-class method using the training data instances and their
	 *  labels.
	 * Depending on the class that implements this method, the algorithm vary,
	 *  but each algorithm will use multiple OnlineEngines and train them
	 * 
	 * @param trainingData the data instances from the training data
	 * @param labelsData the labels for the data instances
	 * @param iterations the number of iterations to train each OnlineEngine
	 *  needed
	 * @param cacheMatrix a boolean that defines the use of a Cache based on a matrix
	 * @param engineObservers a list of EngineOverver objects that will receive 
	 * information in each step of the algorithm that trains the OnlineEngines
	 */
	public abstract void train(final List<DataObject<E>> trainingData,
			final List<L> labelsData, final int iterations, 
			boolean cacheMatrix, 
			final List<EngineObserver<E>> engineObservers);
		
	
	/**
	 * Classifies a data instance with a label
	 * 
	 * @param instance the data instance object
	 * @return the label predicted to the data instance
	 */
	public abstract L getResult(DataObject<E> instance);



}

