package pt.utl.ist.online.learning.multiclass;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import com.davidsoergel.conja.Function;
import com.davidsoergel.conja.Parallel;

import pt.utl.ist.online.learning.KernelBinaryOnlineAlgorithm;
import pt.utl.ist.online.learning.engines.OnlineEngine;
import pt.utl.ist.online.learning.engines.info.EngineObserver;
import pt.utl.ist.online.learning.exceptions.InvalidVectorIndexException;
import pt.utl.ist.online.learning.utils.DataObject;

/**
 * Implements the One Vs All heuristic used for multi-class classification problems.
 * <p>
 * The One Vs All heuristic uses multiple OnlineEngines, one for each possible
 * pair of labels. It then trains each OnlineEngine with the training data
 * instances that belong to the pair of labels assigned to the OnlineEngine.
 * To train, each classifiers votes for one of two labels, in the end the most 
 * voted label will be the answer.
 * 
 * @author Joao L. M. Pereira
 * @version 0.1
 * @since 2015-06-18
 * 
 * @param <E> the class of the data sample
 * @param <L> the class of the labels for each data sample
 */
public class OneVsAllMulticlassMethod<E,L extends Comparable<L>> extends MulticlassMethod<E,L> {
	Map<L,OnlineEngine<E>> classifiers;

/**
 * Creates a multi-class method based on the One Vs all heuristic
 * 
 * @param conf the OnlineEngine.Conf object that can create OnlineEngines
 */
	public OneVsAllMulticlassMethod(OnlineEngine.Conf<E> conf){
		super(conf);
		classifiers = new HashMap<L,OnlineEngine<E>>();
	}

	@Override
	public void train(final List<DataObject<E>> trainingData,
			final List<L> labelsData, final int iterations, 
			boolean cacheMatrix, 
			final List<EngineObserver<E>> engineObservers){

		Set<L> labels = new HashSet<L>(labelsData);
		int numLabels = labels.size();
		int numClassifiers = numLabels;
		System.out.println("Training " + numClassifiers
				+ " one-vs-all classifiers for " + numLabels + " labels");

		if(cacheMatrix){
			conf.cacheMatrix(trainingData);
		}

		addClassifiersToMap(labels);

		// Conja forEach, used to train each classifier in a different thread
		Parallel.forEach(labels, new Function<L, Void>() {
			@Override
			public Void apply(L label) {
				OnlineEngine<E> engine = classifiers.get(label);

				Map<Integer, Boolean> booleanLabels = new HashMap<Integer, Boolean>();

				for(int i = 0; i < labelsData.size(); i++){
					L label2 = labelsData.get(i);
					if(label.compareTo(label2) == 0){
						booleanLabels.put(i, true);
					}else{
						booleanLabels.put(i, false);
					}
				}


				try {
					KernelBinaryOnlineAlgorithm<E> alg = new KernelBinaryOnlineAlgorithm<E>(trainingData,
							booleanLabels, iterations, engine, false,
							engineObservers, label.toString());
				} catch (InvalidVectorIndexException e) {
					e.printStackTrace();
					System.exit(0);
				}
				return null;
			}
		});
	}


	private void addClassifiersToMap(Set<L> labels){

		for(L label: labels){
			if(classifiers.containsKey(label) == false){
				OnlineEngine<E> engine = conf.createEngine();
				classifiers.put(label, engine);
			}
		}

	}

	@Override
	public L getResult(DataObject<E> instance){
		L label = null;
		double maxScore = Double.MAX_VALUE * -1;
		for( Entry<L, OnlineEngine<E>> entry : classifiers.entrySet()){
			OnlineEngine<E> engine = entry.getValue();
			double score = engine.getScore(instance);
			if(score >= maxScore){
				maxScore = score;
				label = entry.getKey();
			}
		}

		return label;
	}



}

